#include "OpenGL.h"
#include "Loader.h"
#include "Console.h"

map<string, RessourceObject> Loader::LoadedObjects;

int Loader::LoadShaderObject(int Type, const char* shader) {
	int length=strlen(shader), compiled, maxLength;
	GLhandleARB Shader=glCreateShaderObjectARB(Type);
	glShaderSourceARB(Shader, 1, &shader, &length);
	glCompileShaderARB(Shader);
	glGetObjectParameterivARB(Shader, GL_OBJECT_COMPILE_STATUS_ARB, &compiled);
	glGetObjectParameterivARB(Shader, GL_OBJECT_INFO_LOG_LENGTH_ARB, &maxLength);
	if (maxLength>1) {
		GLcharARB* pInfoLog = new GLcharARB[maxLength];
		glGetInfoLogARB(Shader, maxLength, &length, pInfoLog);
		string out=pInfoLog;
		int pos=0;
		while (out.substr(pos,out.length()-1).length() > 90) out.insert(pos+=90, "\n");
		if (out.length()) Console::Output(out+"\n");
		delete[] pInfoLog;
	}
	return compiled ? Shader : 0;
}

int Loader::NewShaderProgram(GLhandleARB vrtShader, GLhandleARB frgShader) {
	GLhandleARB Program=glCreateProgramObjectARB();
	glAttachObjectARB(Program, vrtShader);
	glAttachObjectARB(Program, frgShader);
	glLinkProgramARB(Program);
	glDeleteObjectARB(vrtShader);
	glDeleteObjectARB(frgShader);
	int linked, maxLength, length;
	glGetObjectParameterivARB(Program, GL_OBJECT_LINK_STATUS_ARB, &linked);
	glGetObjectParameterivARB(Program,	GL_OBJECT_INFO_LOG_LENGTH_ARB, &maxLength);
	if (maxLength>1) {
		GLcharARB* pInfoLog = new GLcharARB[maxLength];
		glGetInfoLogARB(Program, maxLength, &length, pInfoLog);
		string out=pInfoLog;
		int pos=0;
		while ((pos=out.find('.', pos))!=string::npos) out.insert(++pos, "\n");
		if (out.length()) Console::Output(out+"\n");
		delete[] pInfoLog;
	}
	return linked ? Program : 0;
}

int Loader::LoadMesh(const char* filename) {return 0;}
	
int Loader::LoadShader(const char* filename) {
	map<string, RessourceObject>::iterator it=LoadedObjects.find(filename);
	if (it!=LoadedObjects.end()) return it->second.index;

	FILE* file=fopen(filename, "r");
	if (!file) return Console::Output((string)"ERROR: Opening file "+filename+" failed\n"), 0;

	char* ShaderSource= new char[128 * 42];
	memset(ShaderSource,0,sizeof(ShaderSource));
	char temp[128]={0};
	unsigned shdtype=0;
	fgets(temp, 128, file);
	strcat(ShaderSource, temp);

	if (!memcmp(temp, "!!ARBfp", 7)) shdtype=GL_FRAGMENT_PROGRAM_ARB;
	else if (!memcmp(temp, "!!ARBvp", 7)) shdtype=GL_VERTEX_PROGRAM_ARB;
	else {
		Console::Output("ERROR: Could not determine type of shader! First line has to start with !!ARB[v|f]p\n");
		return 0;
	}

	while (fgets(temp, 128, file)) strcat(ShaderSource, temp);
	fclose(file);
	
	unsigned ID=0;
	glGenProgramsARB(1, &ID);
	glBindProgramARB(shdtype, ID);
	glProgramStringARB(shdtype, GL_PROGRAM_FORMAT_ASCII_ARB,
                    strlen(ShaderSource), ShaderSource);
	delete[] ShaderSource;
	
	int errnum;
	glGetIntegerv(GL_PROGRAM_ERROR_POSITION_ARB, &errnum);
	if (errnum!=-1) {
		char out[300];
		_snprintf(out, 300, "ERROR: Loading Shader:\n%s\n", glGetString(GL_PROGRAM_ERROR_STRING_ARB));
		Console::Output(out);
	}

	LoadedObjects[filename].index=ID;
	LoadedObjects[filename].type=LOT_SHADER;
	return ID;
}

void Loader::UnloadAll() {
	map<string, RessourceObject>::iterator it;
	for (it=LoadedObjects.begin(); it!=LoadedObjects.end(); ++it) {
		switch (it->second.type) {
		case LOT_SHADER: glDeleteProgramsARB(1, &it->second.index); continue;
		case LOT_MESH: continue;
		}
	}
}
