#include "Material.h"
#include "Console.h"
#include <GL/glu.h>

map<string, unsigned> MaterialManager::Textures;
map<string, unsigned> MaterialManager::Shaders;
bool MaterialManager::TexturesEnabled=1;
bool MaterialManager::ShaderEnabled=0;
bool MaterialManager::BlendingEnabled=0;
float MaterialManager::CurrentColor[4]={1};
unsigned MaterialManager::CurrentBlendMode[2]={0};
unsigned MaterialManager::numUnits=0;
unsigned* MaterialManager::BoundTextures=0;
unsigned MaterialManager::BoundShader=0;

struct ddsHeader{
	unsigned Magic;
	unsigned HeaderSize;
	unsigned DDSFlags;
	unsigned Height;
	unsigned Width;
	unsigned Size;
	unsigned Depth;
	unsigned MipMapCount;
	unsigned Reserved[11];

	unsigned PFSize;
	unsigned PFFlags;
	unsigned FourCC;
	unsigned RGBBitCount;
	unsigned RBitMask;
	unsigned GBitMask;
	unsigned BBitMask;
	unsigned AlphaBitMask;
    
    unsigned Caps1;
	unsigned Caps2;
	unsigned DDSX;
	unsigned Whatever[2];
};

void MaterialManager::Init() {
	int num;
	glGetIntegerv(GL_MAX_TEXTURE_UNITS, &num);
	numUnits=num;
	if (!BoundTextures) BoundTextures=new unsigned[numUnits];
	for (int i=numUnits-1; i>=0; --i) BindTexture(i, 0);
	DisableTextures();
	BindShader(0);
	DisableShader();
	SetColor(1,1,1,1);
	BlendMode(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	DisableBlending();
}

void MaterialManager::Cleanup() {
	map<string, unsigned>::iterator it;
	for (it=Textures.begin(); it!=Textures.end(); ++it) glDeleteTextures(1, &it->second);
	for (it=Shaders.begin(); it!=Shaders.end(); ++it) glDeleteProgramsARB(1, &it->second);
	Textures.clear();
	Shaders.clear();
	SDELETEARY(BoundTextures);
	numUnits=0;
}

void MaterialManager::BindTexture(unsigned unit, unsigned tex) {
	if (!TexturesEnabled) {
		glEnable(GL_TEXTURE_2D);
		TexturesEnabled=1;
	}
	if (BoundTextures[unit]==tex) return;
	if (unit>=numUnits) {
		map<string, unsigned>::iterator it;
		for (it=Textures.begin(); it!=Textures.end(); ++it) 
			if (it->second==tex) break;
		char out[150];
		_snprintf(out, 150, "ERROR: Invalid Texture Unit (Texture %s [%d] to unit %d (max %d)\n",
					it->first.c_str(), tex, unit, numUnits-1);
		Console::Output(out);
		return;
	}
	if (unit) glActiveTexture(GL_TEXTURE0+unit);
	glBindTexture(GL_TEXTURE_2D, tex);
	BoundTextures[unit]=tex;
	if (unit) glActiveTexture(GL_TEXTURE0);
};

int MaterialManager::LoadTGA(FILE* file) {
	unsigned char *imageData;
	int bpp, width,	height;

	unsigned char	TGAcompare[12];
	unsigned char	header[6];
	int	bytesPerPixel;
	unsigned imageSize;
	int	temp;
	int	type=GL_RGBA;

	if (fread(TGAcompare,1,sizeof(TGAcompare),file)!=sizeof(TGAcompare) ||
				fread(header,1,sizeof(header),file)!=sizeof(header)) return 0;

	width  = header[1] * 256 + header[0];
	height = header[3] * 256 + header[2];
    
 	if(width<=0 || height<=0) return 0;

	bpp = header[4];							
	bytesPerPixel = bpp/8;
	imageSize = width*height*bytesPerPixel;
	imageData=(unsigned char*)malloc(imageSize);

	if(!imageData || fread(imageData, 1, imageSize, file)!=imageSize) {
		if(imageData) free(imageData);
		return 0;
	}

	if (bpp>16) {
		for(int i=0; i<int(imageSize); i+=bytesPerPixel) {
			temp=imageData[i];
			imageData[i] = imageData[i + 2];
			imageData[i + 2] = temp;
		}
	}
	
	unsigned char* tmpbuf=new unsigned char[imageSize];
	int lineSz=width*bytesPerPixel;
	for(unsigned i=0; i<imageSize; i+=lineSz) 
		memcpy(tmpbuf+i, imageData+(imageSize-lineSz-i), lineSz);

	if (bpp==24) type=GL_RGB;
	if (bpp==16) type=GL_LUMINANCE_ALPHA;
	if (bpp==8) type=GL_LUMINANCE;
	
	unsigned ID;
	glGenTextures(1, &ID);
	BindTexture(0, ID);
	gluBuild2DMipmaps(GL_TEXTURE_2D, type, width, height, type, GL_UNSIGNED_BYTE, tmpbuf);
	free(imageData);
	delete[] tmpbuf;
	
	return ID;
}

/*
	int Size=header.Size;
	
	if (header.Texture & CUBE) {
		Size=0;
		for (int i=0; i<6; ++i)	if (header.Texture & (CUBE_SIDE*i)) Size+=header.Size;
	}
	if (header.Type & MIPMAP) Size<<1;
	if (header.Texture & CUBE) {
		for (int i=0; i<6; ++i) {
			if (header.Texture & (CUBE_SIDE*i)) {
				if (header.FourCC)
					glCompressedTexImage2DARB(TEXTURE_CUBE_MAP_POSITIVE_X_ARB+i, 0, fmt,
					header.Width, header.Height, 0, header.Size, buf);
				else 
					gluBuild2DMipmaps(TEXTURE_CUBE_MAP_POSITIVE_X_ARB+i, 4, 
					header.Width, header.Height, fmt, GL_UNSIGNED_BYTE, buf);
			}
		}
	}

		
	}
	if (header.Texture & VOLUME)
		gluBuild3DMipmaps(GL_TEXTURE_3D, 4, header.Width, header.Height, header.Depth, fmt, GL_UNSIGNED_BYTE, buf);
*/

int MaterialManager::LoadDDS(FILE* file) {
	ddsHeader header;
	fread(&header, 1, sizeof(header), file);
	
	if (!header.MipMapCount) header.MipMapCount=1;
	if (header.MipMapCount>1) header.Size<<=1;
	if (header.Caps2) header.Size*=6;
	
	unsigned char* buf=(unsigned char*)malloc(header.Size);
	fread(buf, 1, header.Size, file);
	int blocksz=16, fmt=0;

	switch(header.FourCC) {
	case 0x31545844: fmt = GL_COMPRESSED_RGBA_S3TC_DXT1_EXT; blocksz=8; break;
	case 0x33545844: fmt = GL_COMPRESSED_RGBA_S3TC_DXT3_EXT; break;
	case 0x35545844: fmt = GL_COMPRESSED_RGBA_S3TC_DXT5_EXT; break;
	case 0x0: fmt = GL_BGRA; break;
	default: Console::Output("ERROR: Unknown format!\n");
	}
	
	unsigned ID;
	glGenTextures(1, &ID);
	BindTexture(0, ID);
		
	if (!header.FourCC && !header.Caps2) 
		gluBuild2DMipmaps(GL_TEXTURE_2D, 4, header.Width, header.Height, fmt, GL_UNSIGNED_BYTE, buf);
	else if(header.Caps2) {
		glEnable(GL_TEXTURE_CUBE_MAP_ARB);
		glBindTexture(GL_TEXTURE_CUBE_MAP_ARB, ID);
		header.Size/=6;
		glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		glCompressedTexImage2DARB(GL_TEXTURE_CUBE_MAP_NEGATIVE_X_ARB, 0,
				fmt, header.Width, header.Height, 0, header.Size, buf);
		glCompressedTexImage2DARB(GL_TEXTURE_CUBE_MAP_POSITIVE_Z_ARB, 0,
				fmt, header.Width, header.Height, 0, header.Size, buf + header.Size);
		glCompressedTexImage2DARB(GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB, 0,
				fmt, header.Width, header.Height, 0, header.Size, buf + 2*header.Size);
		glCompressedTexImage2DARB(GL_TEXTURE_CUBE_MAP_NEGATIVE_Z_ARB, 0,
				fmt, header.Width, header.Height, 0, header.Size, buf + 3*header.Size);
		glCompressedTexImage2DARB(GL_TEXTURE_CUBE_MAP_NEGATIVE_Y_ARB, 0,
				fmt, header.Width, header.Height, 0, header.Size, buf + 4*header.Size);
		glCompressedTexImage2DARB(GL_TEXTURE_CUBE_MAP_POSITIVE_Y_ARB, 0,
				fmt, header.Width, header.Height, 0, header.Size, buf + 5*header.Size);
		
		glDisable(GL_TEXTURE_CUBE_MAP_ARB);
	}
	else {
		for (int i=0, offset=0; i<header.MipMapCount; ++i) {
			header.Size = ((header.Width+3)/4)*((header.Height+3)/4)*blocksz;
			glCompressedTexImage2DARB(GL_TEXTURE_2D, i, fmt, header.Width, header.Height, 0, header.Size, buf + offset);
			offset += header.Size; 
			if (header.Width>1) header.Width >>= 1;
			if (header.Height>1) header.Height >>= 1;
		}
	}
	free(buf);
	return ID;
}

int MaterialManager::LoadTexture(const char* filename, bool mips, int WrapMode) {
	map<string, unsigned>::iterator it=Textures.find(filename);
	if (it!=Textures.end()) return it->second;

	FILE* file=fopen(filename, "rb");
	if (!file) return Console::Output((string)"ERROR: Opening file "+filename+" failed\n"), 0;

	unsigned ID=0;
	const char* ext=&filename[strlen(filename)-3];
	if (!strcmp(ext, "tga")) ID=LoadTGA(file);
	else if (!strcmp(ext, "dds")) ID=LoadDDS(file);
	else Console::Output((string)"ERROR: Unknown file extension: "+filename+'\n');
	fclose(file);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, WrapMode);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, WrapMode);
	if (mips) glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
	else glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,GL_LINEAR);

	if (ID)	Textures[filename]=ID;
	else Console::Output((string)"ERROR: Loading Texture "+filename+" failed\n");
	return ID;
}

int MaterialManager::LoadShader(const char* filename) {
	map<string, unsigned>::iterator it=Shaders.find(filename);
	if (it!=Shaders.end()) return it->second;

	FILE* file=fopen(filename, "r");
	if (!file) return Console::Output((string)"ERROR: Opening file "+filename+" failed\n"), 0;

	char* ShaderSource= new char[128 * 42];
	memset(ShaderSource,0,sizeof(ShaderSource));
	char temp[128]={0};
	unsigned shdtype=0;
	fgets(temp, 128, file);
	strcat(ShaderSource, temp);

	if (!memcmp(temp, "!!ARBfp", 7)) shdtype=GL_FRAGMENT_PROGRAM_ARB;
	else if (!memcmp(temp, "!!ARBvp", 7)) shdtype=GL_VERTEX_PROGRAM_ARB;
	else {
		Console::Output("ERROR: Could not determine type of shader! First line has to start with !!ARB[v|f]p\n");
		return 0;
	}

	while (fgets(temp, 128, file)) strcat(ShaderSource, temp);
	fclose(file);
	
	unsigned ID=0;
	glGenProgramsARB(1, &ID);
	glBindProgramARB(shdtype, ID);
	glProgramStringARB(shdtype, GL_PROGRAM_FORMAT_ASCII_ARB,
                    strlen(ShaderSource), ShaderSource);
	delete[] ShaderSource;
	
	int errnum;
	glGetIntegerv(GL_PROGRAM_ERROR_POSITION_ARB, &errnum);
	if (errnum!=-1) {
		char out[300];
		_snprintf(out, 300, "ERROR: Loading Shader:\n%s\n", glGetString(GL_PROGRAM_ERROR_STRING_ARB));
		Console::Output(out);
	}

	Shaders[filename]=ID;
	return ID;
}
