#include "ParticleSystem.h"
#include "Material.h"
#include "Loader.h"
#include "ProfileCounter.h"

void ParticleSystem::Init() {
	Cleanup();
	Particles=new Particle[ParticleCount*4];
	PartData=new ParticleData[ParticleCount];
	numParticles=ParticleCount;

	for (unsigned i=0; i<ParticleCount; ++i) {
		PartData[i].Velocity.random(1,1,1);
		PartData[i].Lifetime=Lifetime+ (float)(rand()%1000)/500;
		Particles[i].Pos.random(5.7f,5.7f,5.7f, 0);
		Particles[i].Pos.w=1;
		Particles[i].Col=Vector4(r,g,b,a);
		Particles[i].Size.x=Size;
		Particles[i].Size.y=Size*1.3333f;
	}
}

int ParticleSystem::RenderPointSprites() {
	static ProfileCounter* ps=ProfileCounter::NewCounter("Particle PS", .2f, 4);
	ps->Start();
	
	glEnable(GL_POINT_SPRITE_ARB);
	MaterialManager::BindTexture(0, Texture);
	MaterialManager::BlendMode(GL_DST_ALPHA, GL_ONE);
	glDepthMask(0);
	float maxSize;
	float quadratic[] =  {1.0f, 0.0f, 0.01f};
	glPointParameterfvARB(GL_POINT_DISTANCE_ATTENUATION_ARB, quadratic);
	glPointParameterfARB(GL_POINT_FADE_THRESHOLD_SIZE_ARB, 60.0f);
	glGetFloatv(GL_POINT_SIZE_MAX_ARB, &maxSize);
	glPointParameterfARB(GL_POINT_SIZE_MIN_ARB, 0);
	glPointParameterfARB(GL_POINT_SIZE_MAX_ARB, maxSize);
	glTexEnvf(GL_POINT_SPRITE_ARB, GL_COORD_REPLACE_ARB, 1);

	glPointSize(Particles[0].Size.x*200);
	glBegin(GL_POINTS);
	for (unsigned i=0; i<numParticles; ++i) {
		if (Lifetime>0 && (PartData[i].Lifetime-=dTime) <= 0) {
			PartData[i]=PartData[--numParticles];
			Particles[i--]=Particles[numParticles];
			continue;
		}
		if ((Particles[i].Pos+=(PartData[i].Velocity*dTime)).lensq()>100) {
			PartData[i].Velocity*=-1;
			Particles[i].Pos*=.99f;
		}		
		MaterialManager::SetColor(Particles[i].Col.x, Particles[i].Col.y, Particles[i].Col.z, Particles[i].Col.w);
		glVertex3fv(Particles[i].Pos);
	}
	glEnd();
	
	MaterialManager::DisableBlending();
	glDepthMask(1);
	glDisable(GL_POINT_SPRITE_ARB);
	dTime=0;
	ps->Stop();
	return numParticles;
}

int ParticleSystem::RenderBillboards() {
	static ProfileCounter* ps=ProfileCounter::NewCounter("Particle VP", .2f, 4);
	ps->Start();

	MaterialManager::BindTexture(0, Texture);
	MaterialManager::BlendMode(GL_DST_ALPHA, GL_ONE);
	glDepthMask(0);

	Vector3 pos;
	Vector2 size;
		
	glBegin(GL_QUADS);
	for (unsigned i=0; i<numParticles; ++i) {
		if (Lifetime>0 && (PartData[i].Lifetime-=dTime) <= 0) {
			PartData[i]=PartData[--numParticles];
			Particles[i--]=Particles[numParticles];
			continue;
		}
		if ((Particles[i].Pos+=(PartData[i].Velocity*dTime)).lensq()>100) {
			PartData[i].Velocity*=-1;
			Particles[i].Pos*=.99f;
		}
		if (size.x!=Particles[i].Size.x) glTexCoord2fv(size=Particles[i].Size);
		MaterialManager::SetColor(Particles[i].Col.x, Particles[i].Col.y, Particles[i].Col.z, Particles[i].Col.w);
		glNormal3fv(Particles[i].Pos);
		glVertex2f(-.5f,-.5f);
		glVertex2f(.5f,-.5f);
		glVertex2f(.5f,.5f);
		glVertex2f(-.5f,.5f);
	}
	glEnd();
	MaterialManager::DisableBlending();
	glDepthMask(1);
	dTime=0;
	ps->Stop();
	return numParticles<<2;
}

unsigned ParticleManager::numSystems=1;
unsigned ParticleManager::VertexProg=0;
std::vector<ParticleSystem*> ParticleManager::Systems(100);
bool ParticleManager::UseParticles=1;
bool ParticleManager::UsePointSprites=0;

unsigned ParticleManager::CreateSystem() {
	VertexProg=Loader::LoadShader("Particle.vsh");
	Systems[numSystems]=new ParticleSystem;
	char out[100];
	_snprintf(out, 100, "Particle System [%d] created\n", numSystems);
	Console::Output(out);
	return numSystems++;
}

void ParticleManager::DeleteSystem(unsigned idx) {
	Systems[idx]->Active=0;
	Systems[idx]->Cleanup();
	SDELETE(Systems[idx]);
	char out[100];
	_snprintf(out, 100, "Particle System [%d] deleted\n", idx);
	Console::Output(out);
}

void ParticleManager::SetParticleCount(unsigned i, unsigned num) {
	Systems[i]->ParticleCount=num;
	char out[100];
	_snprintf(out, 100, "Particle System [%d] set to %d particles\n", i, num);
	Console::Output(out);
}

void ParticleManager::SetParticleSize(unsigned i, float size) {
	Systems[i]->Size=size;
	char out[100];
	_snprintf(out, 100, "Particle System [%d] set to size %f\n", i, size);
	Console::Output(out);
}

void ParticleManager::SetParticleColor(unsigned i, 
				float r, float g, float b, float a) {
	Systems[i]->r=r;
	Systems[i]->g=g;
	Systems[i]->b=b;
	Systems[i]->a=a;
	char out[100];
	_snprintf(out, 100, "Particle System [%d] set to color %d/%d/%d/%d\n", i, r,g,b,a);
	Console::Output(out);
}

void ParticleManager::SetParticleTexture(unsigned i, const char* file) {
	Systems[i]->Texture=MaterialManager::LoadTexture(file);
	char out[100];
	_snprintf(out, 100, "Particle System [%d] set to texture %s\n", i, file);
	Console::Output(out);
}

void ParticleManager::SetParticleLifetime(unsigned i, float life) {
	Systems[i]->Lifetime=life;
	char out[100];
	_snprintf(out, 100, "Particle System [%d] set to %f lifetime\n", i, life);
	Console::Output(out);
}

void ParticleManager::InitSystem(unsigned i, unsigned num, float life, float size,
									const char* texture, float r, float g, float b, float a) {
	ParticleSystem& p=*Systems[i];
	p.Texture=MaterialManager::LoadTexture(texture);
	p.ParticleCount=num;
	p.Lifetime=life;
	p.Size=size;
	p.r=r; p.g=g; p.b=b; p.a=a;
	p.Init();
	char out[100];
	_snprintf(out, 100, "Particle System [%d] initialized\n", i);
	Console::Output(out);
}

void ParticleManager::ResetSystem(unsigned i, bool run) {
	Systems[i]->Init();
	Systems[i]->Active=run;
	char out[100];
	_snprintf(out, 100, "Particle System [%d] reset%s\n", i, run?" and enabled":"");
	Console::Output(out);
}

void ParticleManager::EnableSystem(unsigned i, bool t) {
	Systems[i]->Active=t;
	char out[100];
	_snprintf(out, 100, "Particle System [%d] %s\n", i, t?"enabled":"disabled");
	Console::Output(out);
}

void ParticleManager::PauseSystem(unsigned i, bool t) {
	Systems[i]->Paused=t;
	char out[100];
	_snprintf(out, 100, "Particle System [%d] %s\n", i, t?"paused":"unpaused");
	Console::Output(out);
}

void ParticleManager::Update(float dtime) {
	if (!UseParticles) return;
	for (unsigned i=1; i<numSystems; ++i) {
		if (!Systems[i] || !Systems[i]->Active) continue;
		if (!Systems[i]->numParticles) {DeleteSystem(i); continue;}
		if (!Systems[i]->Paused) Systems[i]->dTime+=dtime;
	}
}

void ParticleManager::ListSystems() {
	Console::Output("List of Particle Systems:\n"
		"ID\tActive\tPause\t#Particles\tLifetime\tSize\tTexture\t\tColor\n"
		"----------------------------------------------------------------------------------------------\n");
	for (unsigned i=1; i<numSystems; ++i) {
		if (!Systems[i]) continue;
		ParticleSystem& p=*Systems[i];
		char out[200];
		_snprintf(out, 200, "%.2d\t%s\t\t%s\t\t%.7d\t%.2f\t\t%.2f\t%.2d\t\t\t%1.2f/%1.2f/%1.2f/%1.2f\n",
			i, p.Active?"yes":"yes", p.Paused?"yes":"no",
			p.ParticleCount, p.Lifetime, p.Size, p.Texture,	p.r, p.g, p.b, p.a);
		Console::Output(out);
	}
}

int ParticleManager::Render() {
	if (!UseParticles) return 0;
	int numTriangles=0;
	if (UsePointSprites) {
		for (unsigned i=1; i<numSystems; ++i) {
			if (Systems[i]->Active) numTriangles += Systems[i]->RenderPointSprites();
		}
	}
	else {
		glEnable(GL_VERTEX_PROGRAM_ARB);
		glBindProgramARB(GL_VERTEX_PROGRAM_ARB, VertexProg);
		for (unsigned i=1; i<numSystems; ++i) {
			if (Systems[i] && Systems[i]->Active) numTriangles += Systems[i]->RenderBillboards();
		}
		glDisable(GL_VERTEX_PROGRAM_ARB);
	}
	MaterialManager::SetColor(1,1,1,1);
	return numTriangles;
}

void ParticleManager::Cleanup() {
	for (unsigned i=1; i<numSystems; ++i) SDELETE(Systems[i]);
	Systems.clear();
}
