#ifndef PARTICLESYSTEM_INCLUDED
#define PARTICLESYSTEM_INCLUDED

#include <vector>
#include "Globals.h"
#include "VectorX.h"
#include "Console.h"

struct Particle {
	Vector4 Col;
	Vector4 Pos;
	Vector2 Size;
};

struct ParticleData {
	float Lifetime;
	Vector3 Velocity;
};

class ParticleSystem {
	Particle* Particles;
	ParticleData* PartData;
protected:
	float Lifetime;
	float Size;
	unsigned ParticleCount;
	unsigned Texture;
	float r,g,b,a;
	
	float dTime;
	unsigned numParticles;
	bool Paused;
	bool Active;
	
	friend class ParticleManager;

	void Init();
	void Cleanup() {
		SDELETEARY(Particles);
		SDELETEARY(PartData);
		numParticles=0;
	}

public:
	ParticleSystem()
	:Particles(0), PartData(0), ParticleCount(0), dTime(0), 
	 numParticles(0), Paused(0), Active(0) {}
	~ParticleSystem() {Cleanup();}	
	int RenderPointSprites();
	int RenderBillboards();
	int RenderVBO();
};

class ParticleManager {
private:
	static std::vector<ParticleSystem*> Systems;
	static unsigned numSystems;
	static unsigned VertexProg;
	
public:
	static bool UseParticles;
	static bool UsePointSprites;
	static unsigned CreateSystem();
	static void DeleteSystem(unsigned idx);
	
	static void InitSystem(unsigned i, unsigned num, float life, float size,
						const char* texture, float r, float g, float b, float a);
	static void SetParticleCount(unsigned i, unsigned num);
	static void SetParticleSize(unsigned i, float size);
	static void SetParticleColor(unsigned i,
						float r, float g, float b, float a);
	static void SetParticleLifetime(unsigned i, float life);
	static void SetParticleTexture(unsigned i, const char* file);

	static void EnableSystem(unsigned i, bool t);
	static void PauseSystem(unsigned i, bool t);
	static void ResetSystem(unsigned i, bool run);
	static void Update(float dtime);
	static int Render();
	static void ListSystems();
	static void Cleanup();
};

#endif
