#ifndef PROFILECOUNTER_INCLUDED
#define PROFILECOUNTER_INCLUDED

#include "timer.h"
#include "OpenGL.h"
#include "glfont.h"

#include <string>
#include <vector>

template<class T> T inline MaxAr(T* Ar, unsigned num) {
	T Max=Ar[0];
	for (unsigned i=0; i<num; ++i) if (Ar[i] > Max) Max=Ar[i];
	return Max;
}

template<class T> T inline MinAr(T* Ar, unsigned num) {
	T Min=Ar[0];
	for (unsigned i=0; i<num; ++i) if (Ar[i] < Min) Min=Ar[i];
	return Min;
}

class ProfileCounter {
	static unsigned Overhead;
	
	static volatile inline __int64 Ticks() {
			__asm cpuid
			__asm rdtsc
	}
	static void Init() {Overhead=(unsigned)-(Ticks()-Ticks());}
	
	static unsigned FrameTime;
	static std::vector<ProfileCounter*> Counters;

	RDTimer UpdateTimer;
	unsigned *Values;
	float Interval;
	unsigned curPos, Samples, numValues;
	float Update;
	std::string Name;
	float ScaleX, ScaleY;
	float Seconds;
	__int64 StartTicks, SumTicks;
	unsigned Min, Max, Avg, Sum;

public:
	static bool Profiling;

	ProfileCounter(const char* name, float interval, float secs);
	~ProfileCounter();
	void Print(float x, float y, float scale);
	static ProfileCounter* NewCounter(const char* name, float inter, float secs);
	static void DrawCounters();

	inline void Start() {if (Profiling) StartTicks=Ticks();}

	inline void Stop() {
		if (!Profiling) return;
		SumTicks+=Ticks()-(StartTicks+Overhead);
		Update+=UpdateTimer.Reset();
		++Samples;
		if (Update>=Interval) {
			unsigned val=SumTicks/Samples;
			if (val >= Max) Max=val; else if (Values[curPos]==Max) Max=0;
			if (val <= Min) Min=val; else if (Values[curPos]==Min) Min=0;
			Sum+=val-Values[curPos];
			Avg=Sum/numValues;
			Values[curPos++]=val;

			if (!Min) Min=MinAr(Values, numValues);
			if (!Max) Max=MaxAr(Values, numValues);
			if (curPos>=numValues) curPos=0;

			Update-=Interval;
			SumTicks=0;
			Samples=0;			
			
			ScaleY=1.0f;
			while (Max>ScaleY) ScaleY*=10;
			ScaleY=1.0f/ScaleY;
		}
	}
};

#endif
