#include "SCQuadTree.h"

unsigned SCQuadTree::NodeFromPos(float x, float y) {
	unsigned node=0;
	while (node < firstLeaf) {
		unsigned off=0;
		if (x > Data[node].Px) off|=1;
		if (y > Data[node].Pz) off|=2;
		node=(node<<2)+1+off;
	}
	return node;
}

void SCQuadTree::UpdateNodesUp(unsigned node, float miny, float maxy) {
	SCNode& n=Data[node];
	int Min=n.Py-n.Height;
	int Max=n.Py+n.Height;
	if (Min<=miny && Max>=maxy) return;
	if (miny<Min || !Min) Min=(int)miny;
	if (maxy>Max) {Max=(int)maxy; if (maxy-Max) ++Max;}
	n.Py=(short)(.5f*(Min+Max)+.5f);
	n.Height=(unsigned short)(.5f*(Max-Min));
	n.Radius = (unsigned short)sqrt((double)(2*n.Size*n.Size+n.Height*n.Height))+1;
	
	if (node) UpdateNodesUp((node-1)>>2, Min, Max);
}

void SCQuadTree::Init(unsigned rSize, unsigned LeafSize) {
	numLevels=(unsigned)(log10((double)(rSize<<1))/log10(2.0) - log10((double)LeafSize)/log10(2.0));
	numNodes=0;
	for (unsigned i=0; i<numLevels; ++i) numNodes |= (1<<(i*2));
	firstLeaf=numNodes-(1 << (2*numLevels-2));
	Data=new SCNode[numNodes];
	
	unsigned writepos=0, readpos=0;
	rSize>>=1;
	unsigned limit=1;

	Data[writepos++].Set(0,0, rSize);
	while (writepos < numNodes) {
		if (writepos >= limit) {
			limit=(limit<<2)+1;
			rSize>>=1;
		}
		SCNode& n=Data[readpos++];
		Data[writepos++].Set(n.Px-rSize, n.Pz-rSize, rSize);
		Data[writepos++].Set(n.Px+rSize, n.Pz-rSize, rSize);
		Data[writepos++].Set(n.Px-rSize, n.Pz+rSize, rSize);
		Data[writepos++].Set(n.Px+rSize, n.Pz+rSize, rSize);
	}

	char strg[50];
	_snprintf(strg, 50, "Built QuadTree: Nodes: %d, Levels: %d\n", writepos, numLevels);
	Console::Output(strg);
}
