#ifndef SOUNDS_INCLUDED
#define SOUNDS_INCLUDED

#include <AL/al.h>
#include <AL/alut.h>

class Sound {
	unsigned Buffer;
	static vector<unsigned> Sources;
	static map<string, unsigned> Buffers;
public:
	Sound() {};
	Sound(char* file) {Load(file);}
		
	void Load(char* file);
	void Play(unsigned Source, bool Loop);
	void Pause();
	void Stop(unsigned Source) {
		alSourcei(Source, AL_BUFFER, Buffer);
		alSourceStop(Source);
	}
	static unsigned NewSource();
	static void PlaceSource(unsigned Source, float* Position, float* Velocity=0);
	static void PlaceListener(float* Matrix, float* Velocity=0);
	static void DeleteAll();
};

vector<unsigned> Sound::Sources;
map<string, unsigned> Sound::Buffers;

unsigned Sound::NewSource() {
	unsigned ID;
	alGenSources(1, &ID);
	Sources.push_back(ID);
	return ID;
}

void Sound::Play(unsigned Source, bool Loop) {
	alSourcei(Source, AL_BUFFER, Buffer);
	if (Loop) alSourcei(Source, AL_LOOPING, 1);
	else alSourcei(Source, AL_LOOPING, 0);
	alSourcePlay(Source);
}

void Sound::Load(char* file) {
	map<string,unsigned>::iterator it;
	if ((it=Buffers.find(file))!=Buffers.end()) {Buffer=it->second; return;}
	
	void *data;
	unsigned size;
	alGenBuffers(1, &Buffer);

	if (!strcmp(file+(strlen(file)-3), "wav")) {
		ALenum format;
		ALsizei freq;
		char loop;
		alutLoadWAVFile(file, &format, &data, &size, &freq, &loop);
		alBufferData(Buffer, format, data, size, freq);
		alutUnloadWAV(format, data, size, freq);
	}
	if (!strcmp(file+(strlen(file)-3), "ogg")) {
		//alBufferData(Buffer, AL_FORMAT_VORBIS_EXT, data, size, 44100);
	}
}

void Sound::PlaceSource(unsigned Source, float* Position, float* Velocity) {
	alSourcefv(Source, AL_POSITION, Position);
	if (Velocity) alSourcefv(Source, AL_VELOCITY, Velocity);
}

void Sound::PlaceListener(float* Matrix, float* Velocity) {
	float orient[6]={Matrix[8], Matrix[9], Matrix[10], -Matrix[4], -Matrix[5], -Matrix[6]};
	alListenerfv(AL_ORIENTATION, orient);
	alListenerfv(AL_POSITION, &Matrix[12]);
	if (Velocity) alListenerfv(AL_VELOCITY, Velocity);
}

void Sound::DeleteAll() {
	for (unsigned i=0; i<Sources.size(); ++i) alDeleteSources(1, &Sources[i]);
	Sources.clear();
	map<string, unsigned>::iterator it;
	for (it=Buffers.begin(); it!=Buffers.end(); ++it) alDeleteBuffers(1, &it->second);
	Buffers.clear();
}

#endif
