#include "Terrain.h"

void Terrain::SetBasemap(const char* filename) {
	MaterialManager::ReleaseTexture(BaseTexture);
	BaseTexture=MaterialManager::LoadTexture(filename, 0);
}

void Terrain::Cleanup() {
	SDELETE(VBuffer);
	SDELETE(IBuffer);
	SDELETEARY(Heightmap);
	
	if (Patches) {
		int numSegs=(MapSize-1)/PatchSize;
		for (int pat=0; pat<numSegs*numSegs; ++pat) SDELETEARY(Patches[pat].Error);
		SDELETEARY(Patches);
	}
	for (int i=0; i<numBuffers; ++i) SDELETE(Buffers[i]);
	SDELETEARY(Buffers);
	for (int level=0; level<numLevels; ++level)
		for (int l=0; l<4; ++l) SDELETE(Lvl[level].Link[l].LowLvl);
	SDELETE(Lvl);
	QTree.Clear();
	MaterialManager::ReleaseTexture(BaseTexture);
	MaterialManager::ReleaseTexture(BlendTexture);
	MaterialManager::ReleaseTexture(DetailTexture);
	MaterialManager::ReleaseTexture(SkyCube);
}

int Terrain::BuildLinker(int level, int side, int low, unsigned short* buffer) {
	int PS=PatchSize+1;
	int effective_level_high= (level+1)>>1;
	int effective_level_low= (low+1)>>1;
	if (effective_level_high==effective_level_low) return 0;
	int step_high=1<<effective_level_high, step_low=1<<effective_level_low;
	int start_high, start_low, end_high, end_low;
	int idx_step_high, idx_step_low;
	int sign = (side==0 || side==3) ? -1 : 1;
	int low_to_high;
	
	switch (side) {
	case 0: 
		start_high=(PatchSize-step_high)*PS + PatchSize-step_high;
		start_low=PS*PS -1;
		end_high=step_high*PS + PS-(1+step_high);
		end_low=PatchSize;
		idx_step_high=-PS*step_high;
		idx_step_low=-PS*step_low;
		low_to_high=-step_high;
		break;
	case 1:
		start_high=(PatchSize-step_high)*PS +step_high;
		start_low=PatchSize*PS;
		end_high=(PatchSize-step_high)*PS + PatchSize-step_high;
		end_low=PatchSize*PS +PatchSize;
		idx_step_high=step_high;
		idx_step_low=step_low;
		low_to_high=-step_high*PS;
		break;
	case 2:
		start_high=step_high*PS + step_high;
		start_low=0;
		end_high=(PatchSize-step_high)*PS +step_high;
		end_low=PatchSize*PS;
		idx_step_high=PS*step_high;
		idx_step_low=PS*step_low;
		low_to_high=step_high;
		break;
	case 3:
		start_high=step_high*PS + PatchSize-step_high;
		start_low=PatchSize;
		end_high=step_high*PS + step_high;
		end_low=0;
		idx_step_high=-step_high;
		idx_step_low=-step_low;
		low_to_high=step_high*PS;
		break;
	}

	int num=0;
	for (int iter_low=start_low; iter_low*sign <= end_low*sign; iter_low+=idx_step_low) {
		int loop_start_high=(iter_low+low_to_high)-idx_step_low/2;
		int loop_end_high=(iter_low+low_to_high)+idx_step_low/2;

		if (iter_low*sign<end_low*sign) {
			buffer[num++]=iter_low;
			buffer[num++]=iter_low+low_to_high+idx_step_low/2;
			buffer[num++]=iter_low+idx_step_low;
		}
		
		for (int iter_high=loop_start_high; iter_high*sign < loop_end_high*sign; iter_high+=idx_step_high) {
			if (iter_high*sign<start_high*sign || iter_high*sign>=end_high*sign) continue;
			buffer[num++]=iter_high;
			buffer[num++]=iter_high + idx_step_high;
			buffer[num++]=iter_low;
		}
	}

	Lvl[level].Link[side].LowLvl[(low-level)-1].Size=num;
	return num;
}

void Terrain::CreateIndexBuffers() {
	int BufferSize=0, i;
	for (i=0; i<(numLevels+1)>>1; ++i) BufferSize+=(PatchSize*PatchSize)>>(2*i);
	for (i=1; i<(numLevels+1)>>1; ++i) BufferSize+=(PatchSize*PatchSize)>>(2*i -1);
	/**/BufferSize += 2000;/* todo: Calculate real size for linkers*/
	BufferSize*=16*6*sizeof(unsigned short);
	unsigned short* ib;
	IBuffer = new Buffer(INDEX_BUFFER, BufferSize);
	ib=(unsigned short*)IBuffer->Map();
	Lvl=new Level[numLevels];

	int offset=0;
	for (int level=0; level<numLevels; ++level) {
		for (int variant=0; variant<16; ++variant) {
			int num=0;
			Lvl[level].Body[variant].Offset=offset*sizeof(unsigned short);
			if (0==level%2) {
				int lvl=level/2;
				int numSeg=PatchSize>>lvl;
				int PS=PatchSize+1;
				int step=1<<lvl;
				for (int vtx=0; vtx<numSeg*numSeg; ++vtx) {
					int x=(vtx%numSeg)<<lvl, y=(vtx/numSeg)<<lvl;
					if (x==0 && variant&4) {
						if (!(variant&2)) {
							ib[offset++]=PatchSize*PS;
							ib[offset++]=(PatchSize-step)*PS + step;
							ib[offset++]=PatchSize*PS + step;
							num+=3;
						}
						if (!(variant&8)) {
							ib[offset++]=0;
							ib[offset++]=step;
							ib[offset++]=step*PS + step;
							num+=3;
						}
						continue;
					}
					if (x==PatchSize-step && variant&1) {
						if (!(variant&2)) {
							ib[offset++]=PatchSize*PS + PatchSize;
							ib[offset++]=PatchSize*PS + PatchSize-step;
							ib[offset++]=(PatchSize-step)*PS + PatchSize-step;
							num+=3;
						}
						if (!(variant&8)) {
							ib[offset++]=PatchSize;
							ib[offset++]=step*PS + PatchSize-step;
							ib[offset++]=PatchSize-step;
							num+=3;
						}
						continue;
					}
					if (y==0 && variant&8) {
						if (!(variant&1)) {
							ib[offset++]=PatchSize;
							ib[offset++]=step*PS + PatchSize;
							ib[offset++]=step*PS + PatchSize-step;
							num+=3;
						}
						if (!(variant&4)) {
							ib[offset++]=0;
							ib[offset++]=step*PS + step;
							ib[offset++]=step*PS;
							num+=3;
						}
						continue;
					}
					if (y==PatchSize-step && variant&2) {
						if (!(variant&1)) {
							ib[offset++]=PatchSize*PS + PatchSize;
							ib[offset++]=(PatchSize-step)*PS + PatchSize-step;
							ib[offset++]=(PatchSize-step)*PS + PatchSize;
							num+=3;
						}
						if (!(variant&4)) {
							ib[offset++]=PatchSize*PS;
							ib[offset++]=(PatchSize-step)*PS;
							ib[offset++]=(PatchSize-step)*PS + step;
							num+=3;
						}
						continue;
					}
					if (((x+y)>>lvl)%2) {
						ib[offset++]=PS * (y+step) + x;
						ib[offset++]=PS * y + x+step;
						ib[offset++]=PS * (y+step) + x+step;
				
						ib[offset++]=PS * (y+step) + x;
						ib[offset++]=PS * y + x;
						ib[offset++]=PS * y + x+step;
					}
					else {
						ib[offset++]=PS * (y+step) + x;
						ib[offset++]=PS * y + x;
						ib[offset++]=PS * (y+step) + x+step;

						ib[offset++]=PS * y + x;
						ib[offset++]=PS * y + x+step;
						ib[offset++]=PS * (y+step) + x+step;
					}
					num+=6;
				}
			}
			else {
				int lvl=(level/2)+1;
				int numSeg=PatchSize>>lvl;
				int PS=PatchSize+1;
				int step=1<<lvl;
				int halfstep=step>>1;
				for (int vtx=0; vtx<numSeg*numSeg; ++vtx) {
					int x=(vtx%numSeg)<<lvl, y=(vtx/numSeg)<<lvl;

					if (x==0 && variant&4) {
						if (!(variant&2)) {
							ib[offset++]=PatchSize*PS;
							ib[offset++]=(PatchSize-step)*PS + step;
							ib[offset++]=PatchSize*PS + step;
							num+=3;
						}
						if (!(variant&8)) {
							ib[offset++]=0;
							ib[offset++]=step;
							ib[offset++]=step*PS + step;
							num+=3;
						}
						continue;
					}
					if (x==PatchSize-step && variant&1) {
						if (!(variant&2)) {
							ib[offset++]=PatchSize*PS + PatchSize;
							ib[offset++]=PatchSize*PS + PatchSize-step;
							ib[offset++]=(PatchSize-step)*PS + PatchSize-step;
							num+=3;
						}
						if (!(variant&8)) {
							ib[offset++]=PatchSize;
							ib[offset++]=step*PS + PatchSize-step;
							ib[offset++]=PatchSize-step;
							num+=3;
						}
						continue;
					}
					if (y==0 && variant&8) {
						if (!(variant&1)) {
							ib[offset++]=PatchSize;
							ib[offset++]=step*PS + PatchSize;
							ib[offset++]=step*PS + PatchSize-step;
							num+=3;
						}
						if (!(variant&4)) {
							ib[offset++]=0;
							ib[offset++]=step*PS + step;
							ib[offset++]=step*PS;
							num+=3;
						}
						continue;
					}
					if (y==PatchSize-step && variant&2) {
						if (!(variant&1)) {
							ib[offset++]=PatchSize*PS + PatchSize;
							ib[offset++]=(PatchSize-step)*PS + PatchSize-step;
							ib[offset++]=(PatchSize-step)*PS + PatchSize;
							num+=3;
						}
						if (!(variant&4)) {
							ib[offset++]=PatchSize*PS;
							ib[offset++]=(PatchSize-step)*PS;
							ib[offset++]=(PatchSize-step)*PS + step;
							num+=3;
						}
						continue;
					}
					
					ib[offset++]=PS * (y+halfstep) + x+halfstep;
					ib[offset++]=PS * (y+step) + x+step;
					ib[offset++]=PS * (y+step) + x;
				
					ib[offset++]=PS * (y+halfstep) + x+halfstep;
					ib[offset++]=PS * (y+step) + x;
					ib[offset++]=PS * y + x;
				
					ib[offset++]=PS * (y+halfstep) + x+halfstep;
					ib[offset++]=PS * y + x+step;
					ib[offset++]=PS * (y+step) + x+step;

					ib[offset++]=PS * (y+halfstep) + x+halfstep;
					ib[offset++]=PS * y + x;
					ib[offset++]=PS * y + x+step;
	
					num+=12;
				}
			}
			Lvl[level].Body[variant].Size=num;
		}
		
		for (int l=0; l<4; ++l) {
			Lvl[level].Link[l].LowLvl=new Tile[(numLevels-level)-1];
			for (int low=level+1; low<numLevels; ++low) {
				Lvl[level].Link[l].LowLvl[(low-level)-1].Offset=offset*sizeof(unsigned short);
				offset+=BuildLinker(level, l, low, &ib[offset]);
			}
		}
	}
	IBuffer->Unmap();
}

void Terrain::CalcErrors() {
	int numPtch=(MapSize-1)/PatchSize;

	for (int lvl=1; lvl <= numLevels>>1; ++lvl) {
		int step=1<<lvl;
		int num_steps=PatchSize/step;
		
		for (int pat=0; pat<numPtch*numPtch; ++pat) {
			float avg_error_odd=0;
			float avg_error_even=0;
			for (int sz=0; sz<num_steps; ++sz)
			for (int sx=0; sx<num_steps; ++sx) {
				int left=Patches[pat].x + sx*step;
				int right=left+step;
				int top=Patches[pat].z + sz*step;
				int bottom=top+step;

				float tl=Heightmap[top*MapSize + left];
				float tr=Heightmap[top*MapSize + right];
				float bl=Heightmap[bottom*MapSize + left];
				float br=Heightmap[bottom*MapSize + right];
				float mid=Heightmap[(top+(step>>1))*MapSize + left+(step>>1)];

				for (int iz=0; iz<=step; ++iz)
				for (int ix=0; ix<=step; ++ix) {
					float relx=(float)ix/step;
					float relz=(float)iz/step;
					
					float y=Heightmap[(top+iz)*MapSize + left+ix];

					float height_odd=-1, height_even=-1;

					if (relx<=relz) {
						if (relz<=1-relx) 
							height_odd=tl + relz*(bl-tl) + 2*relx*(mid - .5f*(tl+bl)); //left
						else 
							height_odd=bl + relx*(br-bl) + 2*(1-relz)*(mid - .5f*(bl+br)); //bottom
					}
					else {
						if (relz<=1-relx) 
							height_odd=tl + relx*(tr-tl) + 2*relz*(mid - .5f*(tl+tr)); //top
						else 
							height_odd=tr + relz*(br-tr) + 2*(1-relx)*(mid - .5f*(tr+br)); //right
					}
	
					bool bl2tr=!((sx%2) ^ (sz%2));
	
					if (bl2tr) {
						if (relz<=1.0f-relx) 
							height_even=tl + relx*(tr-tl) + relz*(bl-tl); //topleft
						else 
							height_even=br + (1-relx)*(bl-br) + (1-relz)*(tr-br); //botright
					}
					else {					
						if (relx<=relz) 
							height_even=bl + relx*(br-bl) + (1-relz)*(tl-bl); //botleft
						else 
							height_even=tr + (1-relx)*(tl-tr) + relz*(br-tr); //topright
					}

					float error_even=height_even-y;
					if (error_even<0) error_even=-error_even;
					avg_error_even+=error_even;
					
					float error_odd=height_odd-y;
					if (error_odd<0) error_odd=-error_odd;
					avg_error_odd+=error_odd;
				}
			}
			avg_error_even/=(PatchSize+1)*(PatchSize+1);
			avg_error_odd/=(PatchSize+1)*(PatchSize+1);
			if (avg_error_even > Patches[pat].Error[2*lvl])
					Patches[pat].Error[2*lvl]=avg_error_even;
			if (avg_error_odd > Patches[pat].Error[2*lvl -1])
					Patches[pat].Error[2*lvl -1]=avg_error_odd;
		}
	}
}

void Terrain::CreateTerrain() {
	int PS=PatchSize+1;
	int MS1=MapSize-1;
	int numSegs=MS1/PatchSize;
	Patches = new TerrainPatch[numSegs*numSegs];
	numLevels=(int)(log10((double)PatchSize)/log10(2.0))+1;
	numLevels=2*numLevels -1;

	BaseTexture=MaterialManager::LoadTexture("test.tga", 0);
	BlendTexture=MaterialManager::LoadTexture("blend.tga", 0);
	DetailTexture=MaterialManager::LoadTexture("detail.tga", 1, GL_REPEAT);
	SphereTexture=MaterialManager::LoadTexture("sphere.dds", 0);
	SkyCube=MaterialManager::LoadTexture("sky.dds", 0);
	WaterNormal=MaterialManager::LoadTexture("normal.dds", 0);

	WaterFrgProg=MaterialManager::LoadShader("water.psh");
	WaterVrtProg=Loader::LoadShader("water.vsh");

	float* vb;
	unsigned char* bb;
	if (!Unaligned) {
		VBuffer = new Buffer(VERTEX_BUFFER, 2*PS*PS*sizeof(float));
		if (!VBuffer->VBO) {delete VBuffer; Unaligned=1;}
		else {
			vb=(float*)VBuffer->Map();
			for (int i=0; i<PS*PS; ++i) {
				*vb++=(float)(i%PS); *vb++=(float)(i/PS);
			}
		}
	}
	if (Unaligned) {
		VBuffer = new Buffer(VERTEX_BUFFER, 2*PS*PS);
		bb=(unsigned char*)VBuffer->Map();
		for (int i=0; i<PS*PS; ++i) {*bb++=i%PS; *bb++=i/PS;}
	}
	
	QTree.Init(MS1, PatchSize);
	VBuffer->Unmap();

	int VboChunk=JSApp::Config["VboChunk"];
	if (!VboChunk) VboChunk=512;

	numBuffers=((MS1)<VboChunk) ? 1 : (MS1*MS1)/(VboChunk*VboChunk);
	Buffers=new Buffer*[numBuffers];
	unsigned char** bmap=new unsigned char*[numBuffers];
	float** fmap=new float*[numBuffers];
	unsigned* offsets=new unsigned[numBuffers];
	memset(offsets,0, numBuffers*sizeof(unsigned));

	int Segs=(MS1<VboChunk) ? numSegs : (VboChunk/PatchSize);
	int PtchBufferSize=PS*PS*Segs*Segs;
	if (!Unaligned) PtchBufferSize*=sizeof(float);	

	for (int i=0; i<numBuffers; ++i) {
		Buffers[i]= new Buffer(VERTEX_BUFFER, PtchBufferSize);
		if (Unaligned) bmap[i]=(unsigned char*)Buffers[i]->Map();
		else fmap[i]=(float*)Buffers[i]->Map();
	}

	int moff=(PatchSize-MapSize)/2;
	for (int pat=0; pat<numSegs*numSegs; ++pat) {
		int px=(pat%numSegs)*PatchSize, pz=(pat/numSegs)*PatchSize;

		unsigned node=QTree.NodeFromPos((float)px+moff, (float)pz+moff);
		unsigned patch=node-QTree.firstLeaf;
		
		int buffer=pat/(Segs*Segs);
		Patches[patch].VBO=Buffers[buffer];
		Patches[patch].VBOffset=offsets[buffer];
		offsets[buffer]+=PS*PS;
		if (!Unaligned) Patches[patch].VBOffset*=sizeof(float);

		float miny=(float)Heightmap[pz*MapSize+px]/512;
		float maxy=miny;

		for (int poi=0; poi<PS*PS; ++poi) {
			int x=poi%PS, z=poi/PS;
			float y=(float)Heightmap[(pz+z)*MapSize +px+x]/512;
			if (Unaligned) *(bmap[buffer]++)=(unsigned char)y;
			else *(fmap[buffer]++)=y;
			if (y>maxy) maxy=y;
			if (y<miny) miny=y;
		}

		Patches[patch].Error=new float[numLevels];
		Patches[patch].Error[0]=0;
		Patches[patch].LOD=0;
		Patches[patch].x=px;
		Patches[patch].y=0;
		Patches[patch].z=pz;
		int numX=pat%numSegs, numY=pat/numSegs;
		unsigned adj=QTree.NodeFromPos((float)px+PatchSize+moff, (float)pz+moff)-QTree.firstLeaf;
		Patches[patch].Adj[0]=(numX<numSegs-1) ? &Patches[adj] : 0;
		adj=QTree.NodeFromPos((float)px+moff, (float)pz+PatchSize+moff)-QTree.firstLeaf;
		Patches[patch].Adj[1]=(numY<numSegs-1) ? &Patches[adj] : 0;
		adj=QTree.NodeFromPos((float)px-PatchSize+moff, (float)pz+moff)-QTree.firstLeaf;
		Patches[patch].Adj[2]=(numX>0) ? &Patches[adj] : 0;
		adj=QTree.NodeFromPos((float)px+moff, (float)pz-PatchSize+moff)-QTree.firstLeaf;
		Patches[patch].Adj[3]=(numY>0) ? &Patches[adj] : 0;
		
		QTree.UpdateNodesUp(node, miny, maxy);
	}	

	for (int j=0; j<numBuffers; ++j) Buffers[j]->Unmap();
	delete[] bmap;
	delete[] fmap;

	CreateIndexBuffers();
	CalcErrors();
	SetLOD(LOD);

	char out[50];
	Console::Output("Terrain created\n");
	float MemoryUsage=(float)numBuffers*PS*PS*Segs*Segs;
	if (!Unaligned) MemoryUsage*=sizeof(float);
	float TotalMem=MemoryUsage;
	_snprintf(out, 50, "  Vertex Buffer: %.2fkb\n", MemoryUsage/1024);
	Console::Output(out);	
	MemoryUsage=0;
	for (int n=0; n<numLevels; ++n) MemoryUsage+=(PatchSize*PatchSize)>>(2*n);
	MemoryUsage*=6*sizeof(unsigned short);
	TotalMem+=MemoryUsage;
	_snprintf(out, 50, "  Index Buffers: %.2fkb\n", MemoryUsage/1024);
	Console::Output(out);
	MemoryUsage=(float)sizeof(short)*MapSize*MapSize;
	TotalMem+=MemoryUsage;
	_snprintf(out, 50, "  Heightmap: %.2fkb\n", MemoryUsage/1024);
	Console::Output(out);
	MemoryUsage=(float)sizeof(TerrainPatch)*numSegs*numSegs;
	TotalMem+=MemoryUsage;
	_snprintf(out, 50, "  Patch Array: %.2fkb\n", MemoryUsage/1024);
	Console::Output(out);
	MemoryUsage=(float)QTree.numNodes*sizeof(SCNode)+sizeof(SCQuadTree);
	TotalMem+=MemoryUsage;
	_snprintf(out, 50, "  Quadtree: %.2fkb\n", MemoryUsage/1024);
	Console::Output(out);
	_snprintf(out, 50, "  Total Memory used: %.2fkb\n", TotalMem/1024);
	Console::Output(out);
}

void Terrain::NewHeightmap(int msize, int psize) {
	if (!Available) return;
	glFinish();
	Cleanup();
	MapSize=msize+1;
	PatchSize=psize;
	if (PatchSize<=1) {
		PatchSize=2;
		Console::Output("ERROR: PatchSize too small, corrected to 2!\n");
	}
	Heightmap=new unsigned short[MapSize*MapSize];
}

void Terrain::CreateHeightmap(int msize, int psize) {
	Console::Output((string)"Creating Terrain...\n");
	NewHeightmap(msize, psize);
	
	srand((unsigned)RDTimer::GetCPUTicks());

	int i,j,ni,nj,mi,mj,pmi,pmj,rectSize = msize;
	float dh=250*(float)rectSize;

	unsigned short* field=Heightmap;
	field[0] = 0;

	memset(field, 0, MapSize*MapSize);

	while(rectSize > 0)	{		       
		for (i=0;i<msize;i+=rectSize)
		for (j=0;j<msize;j+=rectSize) {
			ni = (i+rectSize)%msize;
			nj = (j+rectSize)%msize;
			mi = (i+rectSize/2);
			mj = (j+rectSize/2);

			field[mi+mj*msize] = (unsigned short)((field[i+j*msize] + field[ni+j*msize] + field[i+nj*msize] + field[ni+nj*msize])/4 -dh/2 + dh*((float)rand()/(float)RAND_MAX));
		}

		for (i=0;i<msize;i+=rectSize)
		for (j=0;j<msize;j+=rectSize) {
			ni = (i+rectSize)%msize;
			nj = (j+rectSize)%msize;
			mi = (i+rectSize/2);
			mj = (j+rectSize/2);
			pmi = (i-rectSize/2+msize)%msize;
			pmj = (j-rectSize/2+msize)%msize;

			field[mi+j*msize] = (unsigned short)((field[i+j*msize] + field[ni+j*msize] + field[mi+pmj*msize] + field[mi+mj*msize])/4 -dh/2 + dh*((float)rand()/(float)RAND_MAX));
			field[i+mj*msize] = (unsigned short)((field[i+j*msize] + field[i+nj*msize] + field[pmi+mj*msize] + field[mi+mj*msize])/4 -dh/2 + dh*((float)rand()/(float)RAND_MAX));
		}
		rectSize /= 2;
		dh /= 2;
	}
	CreateTerrain();
}

void Terrain::LoadHeightmap(const char* file, int msize, int psize) {
	if (!Available) return;
	FILE* fp=fopen(file, "rb");
	if (!fp) {Console::Output((string)"ERROR: Could not load heightmap "+file+"!\n"); return;}
	NewHeightmap(msize, psize);
	Console::Output((string)"Loading Terrain "+file+"...\n");
		
	fread(Heightmap, sizeof(unsigned short), MapSize*MapSize, fp);
	fclose(fp);

	CreateTerrain();
}

float Terrain::GetHeight(float x, float y, int radius) {
	x+=MapSize>>1; y+=MapSize>>1;
	if (x<0 || y<0 || x>MapSize || y>MapSize) return 0;
	int left=(x!=MapSize-1) ? (int)x : (int)x-1;
	int right=left+1;
	int top=(y!=MapSize-1) ? (int)y : (int)y-1;
	int bottom=top+1;
	float relx=x-left;
	float relz=y-top;

	float dx=Heightmap[top*MapSize + left]-Heightmap[top*MapSize + right];
	float dz=Heightmap[bottom*MapSize + left]-Heightmap[top*MapSize + left];
	float height = Heightmap[top*MapSize + left] + relx*dx + relz*dz;
	
	left=(int)(x+.5f); top=(int)(y+.5f);
	for (int itz=(radius>top) ? 0 : top-radius; itz <= (top+radius) && itz<MapSize; itz++)
	for (int itx=(radius>left) ? 0 : left-radius; itx <= (left+radius) && itx<MapSize; itx++)
		if (Heightmap[itz*MapSize + itx]>height) height=Heightmap[itz*MapSize + itx];
	return height/512;
}

void Terrain::SetLoDThreshold(float t) {
	if (t>=0) LodThreshold=t;
	else Console::Output("Threshold must be positive or zero!\n");
}
