#include "Terrain.h"

void Terrain::CreateRenderQ(unsigned node) {
	float center[3]={QTree.Data[node].Px, QTree.Data[node].Py, QTree.Data[node].Pz};
	int vis=Camera.SphereVisible(center, QTree.Data[node].Radius);
	if (node<QTree.firstLeaf) {
		if (vis==1) {
			int num=QTree.GetLeaves(node);
			for (int i=0; i<num; ++i) AddPatch2Q(&Patches[node+i-QTree.firstLeaf]);
		}
		else if (vis==-1) {
			(node<<=2)++;
			CreateRenderQ(node);
			CreateRenderQ(node+1);
			CreateRenderQ(node+2);
			CreateRenderQ(node+3);
		}
	}
	else if (vis) AddPatch2Q(&Patches[node-QTree.firstLeaf]);
}

int Terrain::DrawTree() {
	int numTriangles=0;
	float CamMat[16];
	glGetFloatv(GL_MODELVIEW_MATRIX, CamMat);
	if (Globals::ShowSpheres) {
		
		glEnable(GL_ALPHA_TEST);
		glAlphaFunc(GL_GREATER, .5f);
		MaterialManager::BindTexture(0, SphereTexture);
		float level=-1;
		for (unsigned i=0, lim=0; i<QTree.numNodes; ++i) {
			if (i>=lim) {lim=(lim<<2)+1; ++level;}
			SCNode& n=QTree.Data[i];
			MaterialManager::SetColor(level/QTree.numLevels,
				             .5f*(level/QTree.numLevels), 1-level/QTree.numLevels, 1);

			glPushMatrix();
			float Mat[16]={CamMat[0], CamMat[4], CamMat[8], 0,
							CamMat[1], CamMat[5], CamMat[9], 0,
							CamMat[2], CamMat[6], CamMat[10], 0,
							n.Px, n.Py, n.Pz, 1};
			glMultMatrixf(Mat);

			glBegin(GL_QUADS);
			glTexCoord2f(0,1); glVertex3i(-n.Radius, n.Radius, 0);
			glTexCoord2f(0,0); glVertex3i(-n.Radius, -n.Radius, 0);
			glTexCoord2f(1,0); glVertex3i(n.Radius, -n.Radius, 0);
			glTexCoord2f(1,1); glVertex3i(n.Radius, n.Radius, 0);
			glEnd();
			glPopMatrix();
			numTriangles+=2;
		}
		glDisable(GL_ALPHA_TEST);
		MaterialManager::SetColor(1,1,1,1);
	}
	if (Globals::ShowBoxes) {
		MaterialManager::DisableTextures();
		glDisable(GL_ALPHA_TEST);
		float level=-1;
		for (unsigned i=0, lim=0; i<QTree.numNodes; ++i) {
			if (i>=lim) {lim=(lim<<2)+1; ++level;}
			SCNode& n=QTree.Data[i];
			MaterialManager::SetColor(level/QTree.numLevels,
				                 .5f*(level/QTree.numLevels), 1-level/QTree.numLevels, .33f);
		
			glBegin(GL_LINE_STRIP);
			glVertex3i(n.Px-n.Size, n.Py-n.Height, n.Pz-n.Size);
			glVertex3i(n.Px-n.Size, n.Py-n.Height, n.Pz+n.Size);
			glVertex3i(n.Px-n.Size, n.Py+n.Height, n.Pz+n.Size);
			glVertex3i(n.Px-n.Size, n.Py+n.Height, n.Pz-n.Size);
			glVertex3i(n.Px+n.Size, n.Py+n.Height, n.Pz-n.Size);
			glVertex3i(n.Px+n.Size, n.Py+n.Height, n.Pz+n.Size);
			glVertex3i(n.Px+n.Size, n.Py-n.Height, n.Pz+n.Size);
			glVertex3i(n.Px+n.Size, n.Py-n.Height, n.Pz-n.Size);
			glVertex3i(n.Px-n.Size, n.Py-n.Height, n.Pz-n.Size);
			glVertex3i(n.Px-n.Size, n.Py+n.Height, n.Pz-n.Size);
			glEnd();

			glBegin(GL_LINES);
			glVertex3i(n.Px+n.Size, n.Py+n.Height, n.Pz-n.Size);
			glVertex3i(n.Px+n.Size, n.Py-n.Height, n.Pz-n.Size);
			glVertex3i(n.Px+n.Size, n.Py-n.Height, n.Pz+n.Size);
			glVertex3i(n.Px-n.Size, n.Py-n.Height, n.Pz+n.Size);
			glVertex3i(n.Px-n.Size, n.Py+n.Height, n.Pz+n.Size);
			glVertex3i(n.Px+n.Size, n.Py+n.Height, n.Pz+n.Size);
			glEnd();
		}
		MaterialManager::SetColor(1,1,1,1);
	}
	return numTriangles;
}

void Terrain::SetLOD(int l) {
	if (!Available) return;
	char out[50];
	if (l>=numLevels) {
		_snprintf(out, 50, "Invalid LoD (max %d)\n", numLevels-1);
		Console::Output(out);
		l=numLevels-1;
	}
	if (l<0) l=0;
	_snprintf(out, 50, "LoD set to %d\n", LOD=l);
	Console::Output(out);
}

#include "MatrixX.h"

int Terrain::Render() {
	if (!Available) return 0;
	int numTriangles=0;

	glEnable(GL_TEXTURE_CUBE_MAP_ARB);
	glDisable(GL_DEPTH_TEST);
	glBindTexture(GL_TEXTURE_CUBE_MAP_ARB, SkyCube);
		
	glPushMatrix();
	glMultMatrixf(Camera.Right);
	glMatrixMode(GL_TEXTURE);
	glLoadMatrixf(Camera.Right);
	
	
	glBegin(GL_QUADS);
		glTexCoord4f(-Camera.xFac, +Camera.yFac, 1, 0);
		glVertex3f(-Camera.xFac, +Camera.yFac, 1);
		glTexCoord4f(-Camera.xFac, -Camera.yFac, 1, 0);
		glVertex3f(-Camera.xFac, -Camera.yFac, 1);
		glTexCoord4f(+Camera.xFac, -Camera.yFac, 1, 0);
		glVertex3f(+Camera.xFac, -Camera.yFac, 1);
		glTexCoord4f(+Camera.xFac, +Camera.yFac, 1, 0);
		glVertex3f(+Camera.xFac, +Camera.yFac, 1);
	glEnd();
	glEnable(GL_DEPTH_TEST);
	numTriangles+=6;

	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

	glPushMatrix();
	glEnable(GL_VERTEX_PROGRAM_ARB);

	glBindProgramARB(GL_VERTEX_PROGRAM_ARB, VertexProg);
	glEnableVertexAttribArrayARB(0);
    glEnableVertexAttribArrayARB(1);
	glProgramLocalParameter4fARB(GL_VERTEX_PROGRAM_ARB, 1, 1.0f/(MapSize-1), 8.0f/PatchSize, 10.0f/(Camera.FarP*Camera.FarP), 1);
	
	glTranslatef(-(float)(MapSize-1)/2, 0, -(float)(MapSize-1)/2);
	int Type=Unaligned ? GL_UNSIGNED_BYTE : GL_FLOAT;
	int GeoType=GL_TRIANGLES;
		
	glVertexAttribPointerARB(0, 2, Type, false, 0, VBuffer->Address(0));

	if (Globals::Wireframe) {
		if (JSApp::Config["WireframeAsLines"]) GeoType=GL_LINES;
		else {
			glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
			glDisable(GL_CULL_FACE);
		}
		MaterialManager::DisableTextures();
		MaterialManager::SetColor(1,1,1,1);
	}
	else {
		MaterialManager::BindShader(FragmentProg);
		MaterialManager::BindTexture(2, DetailTexture);
		MaterialManager::BindTexture(1, BlendTexture);
		MaterialManager::BindTexture(0, BaseTexture);
	}

	float fogcolor[]={.53,.75,.99,1};
	glFogfv(GL_FOG_COLOR, fogcolor);

	while (RenderQ) {
		TerrainPatch& p=*RenderQ;
		Level& lvl=Lvl[p.LOD];

		unsigned var=0;
		if (p.Adj[0] && ((p.Adj[0]->LOD+1)>>1 > (p.LOD+1)>>1)) var |= 1;
		if (p.Adj[1] && ((p.Adj[1]->LOD+1)>>1 > (p.LOD+1)>>1)) var |= 2;
		if (p.Adj[2] && ((p.Adj[2]->LOD+1)>>1 > (p.LOD+1)>>1)) var |= 4;
		if (p.Adj[3] && ((p.Adj[3]->LOD+1)>>1 > (p.LOD+1)>>1)) var |= 8;
	
		int maxIndex=(PatchSize+1)*(PatchSize+1) -1;
		glProgramLocalParameter4fARB(GL_VERTEX_PROGRAM_ARB, 0, p.x, p.y, p.z, 0);
		glVertexAttribPointerARB(1, 1, Type, false, 0, p.VBO->Address(p.VBOffset));
		glDrawRangeElements(GeoType, 0, maxIndex, lvl.Body[var].Size, GL_UNSIGNED_SHORT, IBuffer->Address(lvl.Body[var].Offset));
		numTriangles+=lvl.Body[var].Size;
		if (var&1) {
			int lod_diff=(p.Adj[0]->LOD-p.LOD)-1;
			glDrawRangeElements(GeoType,  0, maxIndex, Lvl[p.LOD].Link[0].LowLvl[lod_diff].Size, GL_UNSIGNED_SHORT, 
					IBuffer->Address(Lvl[p.LOD].Link[0].LowLvl[lod_diff].Offset));
			numTriangles+=Lvl[p.LOD].Link[0].LowLvl[lod_diff].Size;
		}
		if (var&2) {
			int lod_diff=(p.Adj[1]->LOD-p.LOD)-1;
			glDrawRangeElements(GeoType, 0, maxIndex, Lvl[p.LOD].Link[1].LowLvl[lod_diff].Size, GL_UNSIGNED_SHORT, 
					IBuffer->Address(Lvl[p.LOD].Link[1].LowLvl[lod_diff].Offset));
			numTriangles+=Lvl[p.LOD].Link[1].LowLvl[lod_diff].Size;
		}
		if (var&4) {
			int lod_diff=(p.Adj[2]->LOD-p.LOD)-1;
			glDrawRangeElements(GeoType, 0, maxIndex, Lvl[p.LOD].Link[2].LowLvl[lod_diff].Size, GL_UNSIGNED_SHORT, 
					IBuffer->Address(Lvl[p.LOD].Link[2].LowLvl[lod_diff].Offset));
			numTriangles+=Lvl[p.LOD].Link[2].LowLvl[lod_diff].Size;
		}
		if (var&8) {
			int lod_diff=(p.Adj[3]->LOD-p.LOD)-1;
			glDrawRangeElements(GeoType, 0, maxIndex, Lvl[p.LOD].Link[3].LowLvl[lod_diff].Size, GL_UNSIGNED_SHORT, 
					IBuffer->Address(Lvl[p.LOD].Link[3].LowLvl[lod_diff].Offset));
			numTriangles+=Lvl[p.LOD].Link[3].LowLvl[lod_diff].Size;
		}
		RenderQ=RenderQ->next;
	}
	Buffer::Unbind(VERTEX_BUFFER);
	Buffer::Unbind(INDEX_BUFFER);
	glPopMatrix();
	if (!JSApp::Config["WireframeAsLines"]) {
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		glEnable(GL_CULL_FACE);
	}

	if (DrawWater) {
		MaterialManager::BlendMode(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		MaterialManager::BindTexture(1, WaterNormal);
		MaterialManager::BindShader(WaterFrgProg);
		glBindProgramARB(GL_VERTEX_PROGRAM_ARB, WaterVrtProg);
		glProgramLocalParameter4fvARB(GL_FRAGMENT_PROGRAM_ARB, 0, Camera.Position);
		glProgramLocalParameter4fvARB(GL_VERTEX_PROGRAM_ARB, 0, Camera.Position);
		glDisable(GL_CULL_FACE);
		glBegin(GL_QUADS);
		  glTexCoord2f(0,0);
		  glVertex3f(-5000, 8.15f, -5000);
		  glTexCoord2f(1,0);
		  glVertex3f( 5000, 8.15f, -5000);
		  glTexCoord2f(1,1);
		  glVertex3f( 5000, 8.15f,  5000);
		  glTexCoord2f(0,1);
		  glVertex3f(-5000, 8.15f,  5000);
		glEnd();
		glEnable(GL_CULL_FACE);
		MaterialManager::DisableBlending();
		numTriangles+=6;
	}

	MaterialManager::DisableShader();
	glDisable(GL_VERTEX_PROGRAM_ARB);
	glDisableVertexAttribArrayARB(0);
    glDisableVertexAttribArrayARB(1);
	glDisable(GL_TEXTURE_CUBE_MAP_ARB);
	if (Globals::ShowBoxes || Globals::ShowSpheres) numTriangles+=DrawTree();

	MaterialManager::DisableTextures();
	glPushMatrix();
	float hf=GetHeight(30.0,270.0-1.96f);
	float hb=GetHeight(30.0,270.0+1.96f);
	float hl=GetHeight(30.0-.84f,270.0);
	float hr=GetHeight(30.0+.84f,270.0);

	Vector3 fwrd(0,hb-hf,2*1.96);
	fwrd.normalize();
	Vector3 right(2*.84f,hr-hl,0);
	right.normalize();
	Vector3 up=fwrd%right;
	Vector3 pos(30, GetHeight(30,270), 270);
	Matrix44 trans(right.x, right.y, right.z, 0,
				   up.x, up.y, up.z, 0,
				   fwrd.x, fwrd.y, fwrd.z, 0,
				   pos.x, pos.y, pos.z, 1);
	glMultMatrixf(trans);
	glBegin(GL_QUAD_STRIP);
	  glVertex3f(-.84f, 1.36f, -1.96f);
	  glVertex3f(-.84f, .2f,   -1.96f);
	  
	  glVertex3f(+.84f, 1.36f, -1.96f);
	  glVertex3f(+.84f, .2f,   -1.96f);
	  
	  glVertex3f(+.84f, 1.36f, +1.96f);
	  glVertex3f(+.84f, .2f,   +1.96f);

	  glVertex3f(-.84f, 1.36f, +1.96f);
	  glVertex3f(-.84f, .2f,   +1.96f);

	  glVertex3f(-.84f, 1.36f, -1.96f);
	  glVertex3f(-.84f, .2f,   -1.96f);
	glEnd();
	glBegin(GL_QUADS);
      glVertex3f(+.84f, 1.36f, +1.96f);
	  glVertex3f(-.84f, 1.36f, +1.96f);
	  glVertex3f(-.84f, 1.36f, -1.96f);
	  glVertex3f(+.84f, 1.36f, -1.96f);
	glEnd();
	glPopMatrix();

	return numTriangles/3;
}
