#ifndef VBUFFER
#define VBUFFER

#include "JSFramework.h"
#include "Console.h"

enum BUFFER_TYPE {VERTEX_BUFFER=0, INDEX_BUFFER};

class Buffer {
private:
	unsigned ID;
	void* Array;
	int GLType;
	int Type;

	static Buffer* Bound[2];

	inline void intBind() {
		if (VBO && Bound[Type]!=this) glBindBufferARB(GLType, (Bound[Type]=this)->ID);
		else if (!VBO && Bound[Type] && Bound[Type]!=this && Bound[Type]->VBO)
			glBindBufferARB(GLType, 0);
	}

public:
	bool VBO;
	void* Map() {
		intBind();		
		return (VBO) ? glMapBufferARB(GLType, GL_WRITE_ONLY_ARB) : Array;
	}
	void Unmap() {
		if (!VBO) return;
		intBind();
		glUnmapBufferARB(GLType);
	}
	void Bind() {
		intBind();
	}
	static void Unbind(BUFFER_TYPE buf) {glBindBufferARB(buf, 0);}
	
	void* Address(unsigned Offset) {
		intBind();
		return (VBO) ? (void*)Offset : (char*)Array+Offset;
	}

	Buffer(BUFFER_TYPE type, unsigned size) : ID(0), Array(0), GLType(0x8892+type), Type(type) {
		VBO=JSApp::Config["UseVBO"] ? 1 : 0;
		if (VBO) {
			glGetError();
			glGenBuffersARB(1, &ID);
			intBind();
			glBufferDataARB(GLType, size, 0, GL_STATIC_DRAW_ARB);
			if (glGetError()!=GL_NO_ERROR) {
				Console::Output("ERROR: Could not create Vertex Buffer in system memory! Falling back to Vertex Array.\n");
				glDeleteBuffersARB(1, &ID);
				VBO=0;
			}
		}
		if (!VBO) {
			if (!(Array=malloc(size))) Console::Output("ERROR: Could not create Vertex Array! Out of memory.\n");
		}
	}
	~Buffer() {
		if (ID) glDeleteBuffersARB(1, &ID);
		if (Array) free(Array);
	}
};

#endif
