#include "JSFramework.h"
#include "ProfileCounter.h"
#include "Collision.h"
#include "Terrain.h"
#include "Loader.h"
#include "Animation2D.h"
#include "Sounds.h"
#include "ParticleSystem.h"
#include "MatrixX.h"
#include "Material.h"

Terrain* terrain=0;

int main(int argc, char *argv[]) {
	JSApp::Init();

	Animation2D Animation("ani.dds", 8,4, 128, 69, 1, 29, 15);

	unsigned SoundSource=Sound::NewSource();
	Sound Engine("idle.wav");
	Engine.Play(SoundSource, 1);

	bool bumm=0;
	Vector3 A, V, N(0,0,0), P(0,0,5), C, newV, a,v;
	float r=0.1f;

	Matrix44 m;
	m.SetIdentity();

	while(!JSApp::Quit) {
		static ProfileCounter* cf=ProfileCounter::NewCounter("Frame Time", .2f, 5);
		cf->Start();
		if (terrain) {
			int limit=terrain->MapSize>>1;
			if (JSApp::Camera.Position[0]< -limit) JSApp::Camera.Position[0]=-limit;
			if (JSApp::Camera.Position[0]>  limit) JSApp::Camera.Position[0]= limit;
			if (JSApp::Camera.Position[2]< -limit) JSApp::Camera.Position[2]=-limit;
			if (JSApp::Camera.Position[2]>  limit) JSApp::Camera.Position[2]= limit;
			float height=1.0f+terrain->GetHeight(JSApp::Camera.Position[0], JSApp::Camera.Position[2], 1);
			if (JSApp::Camera.Position[1]<height) JSApp::Camera.Position[1]=height;
		}
		JSApp::Tick();
		
		float dtime=JSApp::FrameTime;
				
		static ProfileCounter* tp=ProfileCounter::NewCounter("Terrain Culling", .2f, 4);
		tp->Start();
		if (terrain) terrain->Update(dtime);
		tp->Stop();
		Animation.Update(dtime);
		ParticleManager::Update(dtime);

		/*if (!JSApp::Paused) {
			float rad=(float)(rand()%1000)/100000.0f-.005f;
			static ProfileCounter* cx=ProfileCounter::NewCounter("Matrix Mult Sft", .2f, 5);
			cx->Start();
			m.RotateX(rad);
			m.RotateY(rad);
			m.RotateZ(rad);
			cx->Stop();

			char out[150];
			_snprintf(out, 150, "%0.2f %0.2f %0.2f %0.2f\n"
								"%0.2f %0.2f %0.2f %0.2f\n"
								"%0.2f %0.2f %0.2f %0.2f\n"
								"%0.2f %0.2f %0.21 %0.2f\n\n",
				m.X.x, m.X.y, m.X.z, m.X.w,
				m.Y.x, m.Y.y, m.Y.z, m.Y.w,
				m.Z.x, m.Z.y, m.Z.z, m.Z.w,
				m.P.x, m.P.y, m.P.z, m.P.w);
			Console::Output(out);

			double fac=180.0/3.1415927;
			_snprintf(out, 150, "Lengths: %f %f %f\nAngles: %f %f %f\n\n",
				m.X.len(), m.Y.len(), m.Z.len(), acos(m.X*m.Y)*fac, acos(m.X*m.Z)*fac, acos(m.Y*m.Z)*fac);
			Console::Output(out);
		}/**/

		
		JSApp::numTriangles=0;
		if (terrain) JSApp::numTriangles += terrain->Render();
		JSApp::numTriangles += Animation.Render();
		JSApp::numTriangles += ParticleManager::Render();

		Sound::PlaceListener((float*)JSApp::Camera.Right);
		Sound::PlaceSource(SoundSource, A, V);

		if (JSApp::KeyState['A']) {
			N.random(1,1,0); N.normalize();
			A.random(10,10,0);
			V=-A*0.5f; A.z=5;
			C=a=v=0.f; bumm=0;
		}

		if (!JSApp::Paused) {
			Vector3 VelTime=V*dtime;
			float coll=CollisionSpherePlane(A, VelTime, N, P, &C, &newV, r);
			if (coll) {bumm=1; a=C; v=V; A=C+newV*coll; V=newV/dtime;}
			else A+=VelTime;
		}

		Vector3 T=A.operator-(JSApp::Camera.Position);
		glPointSize(r*1050.0/T.len());
				
		MaterialManager::DisableTextures();
		glBegin(GL_POINTS);
		glVertex3fv(A);
		glEnd();
		glPointSize(1);
		
		glBegin(GL_LINES);
		glVertex3f(-5,0,5); glVertex3f(5,0,5);
		glVertex3f(0,-5,5); glVertex3f(0,5,5);
		
		if (bumm) {
			glVertex3fv(a);  glVertex3fv(a+v);
			glVertex3fv(a);  glVertex3fv(a-v);
			glVertex3fv(a);  glVertex3fv(a+V);
		}
		glVertex3fv(P);  glVertex3fv(P+N);
		glVertex3f(4*(P.x-N.y), 4*(P.x+N.x), 5); glVertex3f(4*(P.x+N.y), 4*(P.x-N.x), 5);
		glEnd();
		
		JSApp::ShowStats();
		cf->Stop();
	}

	SDELETE(terrain);
	Loader::UnloadAll();
	Sound::DeleteAll();
	ParticleManager::Cleanup();
	JSApp::Shutdown();
	return 0;
}
